#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

from ansible_anonymizer.jinja2 import str_jinja2_variable_name


def test_str_jinja2_variable_name_leading_underscore():
    assert str_jinja2_variable_name("-foo-BAR") == "foo_bar"



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name
from collections.abc import Generator
from enum import Enum
from typing import Optional, Union

from .field_checks import is_password_field_name


class ParserError(Exception):
    """Unexpected behaviour."""


class NodeType(Enum):
    """The different type of Node returned by the parser."""

    # pylint: disable=invalid-name

    unknown = 0
    field = 1
    separator = 2
    quoted_string_holder = 3
    quoted_string_closing = 4
    new_line = 5
    space = 6
    securized = 7
    backslash = 8
    deleted = 9
    secret = 10


class Node:
    # pylint: disable=too-many-instance-attributes
    """A element returned by the parser."""

    def __init__(self, begin_at: int) -> None:
        self.previous: Optional["Node"] = None
        self.next: Optional["Node"] = None
        self.begin_at: int = begin_at
        self.end_at: int
        self.text: str = ""
        self.type: NodeType = NodeType.unknown
        self.holder: Optional["Node"] = None
        # Node of the field that point on this secret
        self.secret_value_of: Optional["Node"] = None

        # NOTE: Fields only used with quoted strings (called `holder`)
        self.closed_by: Optional["Node"] = None
        self.sub: list["Node"] = []
        self.is_protected: bool = False

    def attach(self, previous: "Node") -> None:
        """Attach a new Node to the previous one in the series."""
        self.previous = previous
        previous.next = self
        holder = previous
        while holder:
            if holder.type is NodeType.quoted_string_holder and not holder.closed_by:
                holder.sub.append(self)
                self.holder = holder
                break
            if not holder.previous:
                break
            holder = holder.previous

    def get_secret(self) -> Union["Node", None]:
        """Identify the secret Node associated with the current Node."""
        candidate = self.next
        has_separator = False
        while candidate:
            if candidate.type is NodeType.space:
                pass
            elif candidate.type is NodeType.quoted_string_closing:
                pass
            elif has_separator:
                if candidate.type is NodeType.securized:
                    # We should not come back on the same node
                    raise ParserError
                if candidate.type in [
                    NodeType.field,
                    NodeType.unknown,
                    NodeType.quoted_string_holder,
                ]:
                    return candidate
                return None
            elif candidate.type is NodeType.separator:
                has_separator = True
            else:
                return None
            candidate = candidate.next
        return None

    def merge_with_next(self) -> None:
        """Merge the current node with the next one."""
        self.type = NodeType.unknown
        assert self.next  # for mypy # noqa: S101
        self.text += self.next.text
        self.next.type = NodeType.deleted
        self.next.text = "KILLED"
        self.next = self.next.next
        if self.next:
            self.next.previous = self

    def is_password_field_name(self) -> bool:
        """Return True if the field name matches the DENYLIST_REGEX regex."""
        return is_password_field_name(self.text)

    def __str__(self) -> str:
        """Expose the Node as a string."""
        return f"TEXT={self.text}, BEGIN_AT={self.begin_at}, TYPE={self.type}"

    def get_previous_nodes(self, limit: int = 0) -> Generator["Node", None, None]:
        """Return the nodes preceding the current one."""
        c = self.previous
        cpt = 0
        while c:
            yield c
            cpt += 1
            if limit and cpt >= limit:
                break
            c = c.previous

    def get_next_nodes(self, limit: int = 0) -> Generator["Node", None, None]:
        """Return the nodes following the current one."""
        c = self.next
        cpt = 0
        while c:
            yield c
            cpt += 1
            if limit and cpt >= limit:
                break
            c = c.next



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

from ansible_anonymizer.node import NodeType
from ansible_anonymizer.parser import parse_raw_block


def test_get_previous():
    sample = """
    "aaa
       \\"passwd: bob\\""
    """
    root_node = parse_raw_block(sample)
    assert [n.type for n in root_node.get_previous_nodes()] == []
    assert [n.type for n in root_node.next.next.next.get_previous_nodes(limit=1)] == [
        NodeType.space
    ]
    assert [n.type for n in root_node.next.next.next.get_previous_nodes()] == [
        NodeType.space,
        NodeType.new_line,
        NodeType.quoted_string_holder,
    ]


def test_get_next():
    sample = """
    "aaa
       \\"passwd: bob\\""
    """
    root_node = parse_raw_block(sample)
    assert [n.type for n in root_node.get_next_nodes()][:3] == [
        NodeType.new_line,
        NodeType.space,
        NodeType.space,
    ]
    assert [n.type for n in root_node.get_next_nodes(limit=1)] == [NodeType.new_line]



#!/usr/bin/env python3
"""Parser for YAML-like structure that is error tolerant."""
from collections.abc import Generator
from typing import Optional, Union

from .node import Node, NodeType
from .parser_multi_lines import group_multi_lines


def is_valid_first_character_for_a_variable(char: str) -> bool:
    """Assuming variable names cannot start with a digit."""
    return char.isascii() and (char.isalpha() or char in ["-", "_"])


def is_valid_variable_character(char: str) -> bool:
    """Return True if the character can be part of an Ansible variable name."""
    # note: isnumeric accepts a wider range than just the 0-9
    return char.isascii() and (char.isalpha() or char.isnumeric() or char in ["-", "_"])


def is_field_value_sep(char: str) -> bool:
    """Return True if the character is : or =."""
    return char in [":", "="]


def breakup_elements(block: str) -> Node:
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements
    """Digest a text block an return a list of Nodes that will be simplified later."""
    root_node = Node(0)
    root_node.type = NodeType.quoted_string_holder
    current_node = root_node
    for pos, c in enumerate(block):  # pylint: disable=invalid-name
        previous_node = current_node
        current_node = Node(-1)  # -1 == undef, the variable will be reset
        if c == "\\":
            new_node = Node(pos)
            new_node.attach(previous=previous_node)
            new_node.type = NodeType.backslash
            current_node = new_node
        elif c in ["'", '"']:
            is_protected = previous_node.type is NodeType.backslash

            holder: Optional[Node] = previous_node
            while holder:
                if (
                    holder.text == c
                    and holder.is_protected is is_protected
                    and not holder.closed_by
                ):
                    break
                holder = holder.holder

            if holder:
                new_node = Node(pos)
                new_node.attach(previous=previous_node)
                new_node.type = NodeType.quoted_string_closing
                holder.closed_by = new_node
                current_node = new_node
            else:
                new_node = Node(pos)
                new_node.type = NodeType.quoted_string_holder
                new_node.is_protected = is_protected
                new_node.attach(previous=previous_node)
                current_node = new_node
        elif is_valid_first_character_for_a_variable(c):
            if previous_node.type is NodeType.field:
                current_node = previous_node
            else:
                new_node = Node(pos)
                new_node.attach(previous=previous_node)
                new_node.type = NodeType.field
                current_node = new_node
        elif is_valid_variable_character(c):
            if previous_node.type is NodeType.field:
                current_node = previous_node
            else:
                new_node = Node(pos)
                new_node.attach(previous=previous_node)
                new_node.type = NodeType.field
                current_node = new_node
        elif is_field_value_sep(c):
            new_node = Node(pos)
            new_node.attach(previous=previous_node)
            new_node.type = NodeType.separator
            current_node = new_node
        elif c == "\n":
            new_node = Node(pos)
            new_node.attach(previous=previous_node)
            new_node.type = NodeType.new_line
            current_node = new_node
        elif c == " ":
            new_node = Node(pos)
            new_node.attach(previous=previous_node)
            new_node.type = NodeType.space
            current_node = new_node
        elif previous_node.type is not NodeType.unknown:
            new_node = Node(pos)
            new_node.attach(previous=previous_node)
            current_node = new_node
        else:
            # Should never happend
            new_node = Node(pos)
            new_node.attach(previous=previous_node)
            current_node = new_node

        current_node.text += c
    return root_node


def parse_raw_block(block: str) -> Node:
    """Return block without any potential secrets."""
    root_node = breakup_elements(block)
    close_quotes(root_node)
    group_multi_lines(root_node)
    handle_backslashes(root_node)
    combinate_value_fields(root_node)
    identify_secrets(root_node)
    return root_node


def close_quotes(root_node: Node) -> None:
    """Close the quotes that are still opened."""
    current_node = root_node.next
    while current_node:
        if current_node.type is NodeType.quoted_string_holder and not current_node.closed_by:
            current_node.type = NodeType.unknown
        current_node = current_node.next


def handle_backslashes(root_node: Node) -> None:
    """
    Convert the backslashes are regular unkown leaves.

    The backslash type is only used to identify the protected quotes.
    """
    current_node = root_node.next
    while current_node:
        if current_node.type is not NodeType.backslash:
            pass
        elif (
            current_node.next
            and current_node.next.type
            in [NodeType.quoted_string_holder, NodeType.quoted_string_closing]
            and current_node.holder is current_node.next.holder
        ):
            current_node.next.text = current_node.text + current_node.next.text
            if current_node.previous:
                current_node.previous.next = current_node.next
                current_node.next.previous = current_node.previous
            current_node.text = "KILLED"
            current_node.type = NodeType.deleted
            current_node = current_node.next
        else:
            current_node.type = NodeType.unknown
        current_node = current_node.next


def combinate_value_fields(root_node: Node) -> None:
    """Combinate the unquoted value strings."""
    mergable_types = [NodeType.space, NodeType.field, NodeType.unknown, NodeType.space]

    def _is_a_new_key_value(node: Node) -> bool:
        """Check if the current node is actually the beginning of a new secret key/value."""
        if not node.next:
            return False

        current = node.next
        if current.type is not NodeType.space:
            return False

        while current.type == NodeType.space:
            if current.next is None:
                return False
            current = current.next

        # if we've got a separator after the field, we prefer to preserve
        # it. e.g: secret1: foo secret2: bar, secret1 and secret2 are two distinct seecrets.
        if current and current.type is NodeType.field and current.next and current.next.text == ":":
            return True
        return False

    def _find_separator_node(node: Node) -> Union[None, Node]:
        while node and node.next:
            if node.next.type is NodeType.space:
                node = node.next
            elif node.next and node.next.type is NodeType.separator:
                return node.next
            else:
                return None
        return None

    current_node = root_node
    while current_node and current_node.next:
        current_node = current_node.next
        if current_node.type is not NodeType.field:
            continue
        if not current_node.is_password_field_name():
            continue

        secret_content_ptr = current_node.get_secret()
        if not secret_content_ptr:
            continue
        separator = _find_separator_node(current_node)
        if not separator:
            continue
        if secret_content_ptr.type is NodeType.quoted_string_holder:
            current_node = secret_content_ptr.closed_by
            continue
        while (
            secret_content_ptr.next
            and (
                (secret_content_ptr.next.type in mergable_types)
                or (
                    secret_content_ptr.next.type is NodeType.separator
                    and secret_content_ptr.next.text != separator.text
                )
            )
            and not _is_a_new_key_value(secret_content_ptr)
        ):
            secret_content_ptr.merge_with_next()


def flatten(node: Node) -> Generator[Node, None, None]:
    """Iterator that go through each nodes and follow the original text order."""
    current = node
    while current:
        yield current
        if not current.next:
            break
        current = current.next


def identify_secrets(root_node: Node) -> None:
    """Remove the secret fields from a series of nodes."""

    def identify_quoted_string(node: Node, secret_node: Node) -> None:
        assert secret_node.closed_by  # for mypy # noqa: S101
        cursor = secret_node.next
        while cursor and cursor != secret_node.closed_by and cursor.next != secret_node.closed_by:
            cursor.merge_with_next()
        if secret_node.next and secret_node.next != secret_node.closed_by:
            secret_node.next.secret_value_of = node
            secret_node.next.type = NodeType.secret
        if secret_node.next:
            identify_secrets(secret_node.next)

    def identify_regular_field(node: Node, secret_node: Node) -> None:
        assert secret_node.holder  # for mypy # noqa: S101
        secret_node.secret_value_of = node
        secret_node.type = NodeType.secret
        if secret_node.next:
            identify_secrets(secret_node.next)

    for node in flatten(root_node):
        if node.type is not NodeType.field:
            continue
        if not node.is_password_field_name():
            continue

        secret_node = node.get_secret()
        if not secret_node:
            continue

        if secret_node.type is NodeType.quoted_string_holder:
            return identify_quoted_string(node, secret_node)
        return identify_regular_field(node, secret_node)



#!/usr/bin/env python3
"""Functions used to identify the field types."""
import re

# Denylist regex to TC of secrets filter
# From detect_secrets.plugins (Apache v2 License)
DENYLIST = (
    "api_?key",
    "auth_?key",
    "service_?key",
    "account_?key",
    "db_?key",
    "database_?key",
    "priv_?key",
    "private_?key",
    "client_?key",
    r"host\w*_key",
    "db_?pass",
    "database_?pass",
    "key_?pass",
    "key_?data",
    "key_?name",
    "password",
    "passwd",
    "pass",
    "pwd",
    "secret",
    "contraseña",
    "contrasena",
    "access_key",
)
AFFIX_REGEX = r"\w*"
DENYLIST_REGEX = r"|".join(DENYLIST)
# Support for suffix after keyword i.e. password_secure = "value"
DENYLIST_REGEX_WITH_PREFIX = fr"({DENYLIST_REGEX}){AFFIX_REGEX}"


def is_allowed_password_field(field_name: str) -> bool:
    """Return True if field_name should not be considered as a password."""
    # Valid field found in sudo configuration
    if field_name == "NOPASSWD":
        return True
    return False


def is_password_field_name(name: str) -> bool:
    """Return True if name looks like a password field name."""
    flags = re.MULTILINE | re.IGNORECASE
    if is_allowed_password_field(name):
        return False
    return re.search(DENYLIST_REGEX_WITH_PREFIX, name, flags=flags) is not None


def is_jinja2_expression(value: str) -> bool:
    """Check if an unquoted string hold a Jinja2 variable."""
    if re.match(r"^\s*{{\s*.*?\s*}}\s*$", value):
        return True

    return False


def is_uuid_string(value: str) -> bool:
    """Check if a given value is a UUID string."""
    if re.match(
        r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
        value,
        flags=re.IGNORECASE,
    ):
        return True

    return False


def is_path(content: str) -> bool:
    """Return True if content is a path."""
    # Rather conservative on purpose to avoid a false
    # positive
    if "/" not in content:
        return False
    return bool(re.match(r"^(|~)[a-z0-9_/\.-]+$", content, flags=re.IGNORECASE))



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
import argparse
import pathlib

import yaml

from ansible_anonymizer.anonymizer import anonymize_struct, anonymize_text_block


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("file_path", type=pathlib.Path)
    parser.add_argument("--format", choices=["text", "yaml"], type=str, default="text")
    args = parser.parse_args()

    if args.format == "text":
        print(anonymize_text_block(args.file_path.read_text()), end="")
    elif args.format == "yaml":
        print(anonymize_struct(yaml.safe_load(args.file_path.read_text())))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name
import ipaddress
import re
from collections.abc import Generator
from ipaddress import IPv4Address, IPv6Address
from re import Match
from string import Template
from typing import Any, Optional
from zlib import crc32

from ansible_anonymizer.field_checks import (
    is_jinja2_expression,
    is_password_field_name,
    is_path,
    is_uuid_string,
)
from ansible_anonymizer.jinja2 import str_jinja2_variable_name
from ansible_anonymizer.parser import flatten, parse_raw_block

from .node import NodeType


def gen_email_address(original: Match[str]) -> str:
    samples = [
        "liam",
        "olivia",
        "noah",
        "emma",
        "oliver",
        "charlotte",
        "elijah",
        "amelia",
        "james",
        "ava",
        "william",
        "sophia",
        "benjamin",
        "isabella",
        "lucas",
        "mia",
        "henry",
        "evelyn",
        "theodore",
        "harper",
    ]
    idx = crc32(original.group("email").encode()) % len(samples)
    name = samples[idx]
    return f"{name}{idx}@example.com"


common_ipv4_networks = [
    ipaddress.IPv4Network("1.0.0.1/32"),
    ipaddress.IPv4Network("1.1.1.1/32"),
    ipaddress.IPv4Network("149.112.112.112/32"),
    ipaddress.IPv4Network("208.67.220.220/32"),
    ipaddress.IPv4Network("208.67.222.222/32"),
    ipaddress.IPv4Network("76.223.122.150/32"),
    ipaddress.IPv4Network("76.76.19.19/32"),
    ipaddress.IPv4Network("8.20.247.20/32"),
    ipaddress.IPv4Network("8.26.56.26/32"),
    ipaddress.IPv4Network("8.8.4.4/32"),
    ipaddress.IPv4Network("8.8.8.8/32"),
    ipaddress.IPv4Network("9.9.9.9/32"),
    ipaddress.IPv4Network("94.140.14.14/32"),
    ipaddress.IPv4Network("94.140.15.15/32"),
    ipaddress.IPv4Network("255.0.0.0/4", False),
    ipaddress.IPv4Network("255.255.255.255/32"),
]


def redact_ipv4_address(value: IPv4Address) -> IPv4Address:
    for i in common_ipv4_networks:
        if value in i:
            return value
    try:
        return value + int(value) % 100
    except ipaddress.AddressValueError:
        return value


common_ipv6_networks = [
    ipaddress.IPv6Network("2001:4860:4860::8888/128"),
    ipaddress.IPv6Network("2001:4860:4860::8844/128"),
]


def redact_ipv6_address(value: IPv6Address) -> IPv6Address:
    for i in common_ipv6_networks:
        if value in i:
            return value

    def randomize(block: Match[str]) -> str:
        field = block.group(0)[1:]
        as_int = int(field, 16)
        new_val = as_int % 1024
        as_hex = hex(new_val)
        hex_without_0x_prefix = as_hex[2:]
        return ":" + hex_without_0x_prefix

    new_address = ipaddress.IPv6Address(
        re.sub(r":[a-z0-9]+", randomize, value.compressed, flags=re.IGNORECASE)
    )
    return new_address


def redact_ip_address(value: str) -> str:
    ip = ipaddress.ip_address(value)
    if ip.version == 4:
        return str(redact_ipv4_address(ip))
    return str(redact_ipv6_address(ip))


def unquote(value: str) -> str:
    if not value or len(value) < 2:
        return value
    if value[0] == value[-1] and value[0] in ('"', "'"):
        return value[1:-1]
    return value


def anonymize_field(value: str, name: str, value_template: Template) -> str:
    v = value.strip()
    if is_uuid_string(v):
        return value
    if is_password_field_name(name):
        if is_path(v):
            return value
        if is_jinja2_expression(unquote(v)):
            return unquote(v)
        variable_name = str_jinja2_variable_name(name)
        return value_template.substitute(variable_name=variable_name)
    return anonymize_text_block(value, value_template=value_template)


def anonymize_struct(o: Any, key_name: str = "", value_template: Optional[Template] = None) -> Any:
    if not value_template:
        value_template = Template("{{ $variable_name }}")

    def key_name_str(k: Any) -> str:
        return k if isinstance(k, str) else ""

    if key_name and not isinstance(key_name, str):
        key_name = str(key_name)

    if isinstance(o, dict):
        return {
            k: anonymize_struct(v, key_name=key_name_str(k), value_template=value_template)
            for k, v in o.items()
        }
    if isinstance(o, list):
        return [anonymize_struct(v, key_name=key_name, value_template=value_template) for v in o]
    if isinstance(o, str):
        return anonymize_field(o, key_name, value_template)
    return o


def anonymize(o: Any, key_name: str = "") -> Any:
    """Deprecated: use anonymize_struct() instead."""
    return anonymize_struct(o, key_name=key_name)


def hide_emails(block: str) -> str:
    flags = re.MULTILINE | re.DOTALL | re.IGNORECASE
    email_re = r"\b\S+@[a-z\.]+[a-z]{2,}\b"
    return re.sub(fr"(?P<email>{email_re})", gen_email_address, block, flags=flags)


def hide_ip_addresses(block: str) -> str:
    flags = re.MULTILINE | re.DOTALL | re.IGNORECASE

    def _rewrite(m: re.Match[str]) -> str:
        try:
            ip = ipaddress.ip_address(m.group("ip_address"))
        except ValueError:
            return m.group("ip_address")
        if ip.version == 4:
            return str(redact_ipv4_address(ip))
        return str(redact_ipv6_address(ip))

    return re.sub(
        r"(?P<ip_address>(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})|[a-f\d:]{3,32})",
        _rewrite,
        block,
        flags=flags,
    )


def hide_us_ssn(block: str) -> str:
    flags = re.MULTILINE | re.DOTALL | re.IGNORECASE

    def _rewrite(_: re.Match[str]) -> str:
        return "{{ ssn }}"

    us_ssn_regex = r"\b(?!666|000|9\d{2})\d{3}-(?!00)\d{2}-(?!0{4})\d{4}\b"
    return re.sub(us_ssn_regex, _rewrite, block, flags=flags)


def hide_mac_addresses(block: str) -> str:
    flags = re.MULTILINE | re.DOTALL | re.IGNORECASE

    def _rewrite(m: re.Match[str]) -> str:
        idx = crc32(m.group("mac").encode())

        def gen() -> Generator[str, None, None]:
            for c in m.group("mac"):
                if c in ["-", ":", "."]:
                    yield c
                else:
                    yield str(hex(int(c, 16) + idx % 0xF)[-1])

        return "".join(c for c in gen())

    mac_regex = (
        r"(?P<mac>\b([0-9a-f]{2}[:-])"
        + r"{5}([0-9a-f]{2})|"  # noqa: W503
        + r"([0-9a-f]{4}\."  # noqa: W503
        + r"[0-9a-f]{4}\."  # noqa: W503
        + r"[0-9a-f]{4})\b)"  # noqa: W503
    )
    return re.sub(mac_regex, _rewrite, block, flags=flags)


def hide_us_phone_numbers(block: str) -> str:
    flags = re.MULTILINE | re.DOTALL | re.IGNORECASE

    def _rewrite(m: re.Match[str]) -> str:
        return m.group("before") + "(311) 555-2368" + m.group("after")

    pattern_before = r"(?P<before>([^\d\.]|^))"
    pattern_after = r"(?P<after>([^\d\.]|$))"
    regexes = [
        r"(?P<number>\d{10})",
        r"(?P<number>1\d{10})",
        r"(?P<number>\d{3}-\d{3}-\d{4})",
        r"(?P<number>\d{3} \d{3}-\d{4})",
        r"(?P<number>\(\d{3}\) \d{3}-\d{4})",
    ]

    for r in regexes:
        full_regex = pattern_before + r + pattern_after
        block = re.sub(full_regex, _rewrite, block, flags=flags)
    return block


def hide_credit_cards(block: str) -> str:
    flags = re.MULTILINE | re.DOTALL | re.IGNORECASE

    def _rewrite(m: re.Match[str]) -> str:
        def luhn(n: str) -> bool:
            r = [int(ch) for ch in str(n)][::-1]
            return (sum(r[0::2]) + sum(sum(divmod(d * 2, 10)) for d in r[1::2])) % 10 == 0

        cc = m.group("cc").replace(" ", "").replace("-", "")
        new_value = "{{ credit_card_number }}" if luhn(cc) else m.group("cc")
        return m.group("before") + new_value + m.group("after")

    cc_regex = r"(?P<before>([^\d-]|^))(?P<cc>(?:\d[ -]*?){13,16})(?P<after>([^\d-]|$))"

    return re.sub(cc_regex, _rewrite, block, flags=flags)


def hide_comments(block: str) -> str:
    new_block = ""
    quotes = ""
    in_comment = False
    for c in block:
        if c == "\n":
            in_comment = False
            quotes = ""
            new_block += c
        elif in_comment:
            continue
        elif c in ['"', "'"]:
            if quotes and quotes[-1] == c:
                quotes = quotes[:-1]
            else:
                quotes += c
            new_block += c
        elif c == "#" and not quotes:
            in_comment = True
            new_block = new_block.rstrip(" ")
        else:
            new_block += c
    return new_block


def hide_user_name(block: str) -> str:
    flags = re.IGNORECASE

    known_users = {
        "cloud-user",
        "ec2-user",
        "fedora",
        "root",
        "ubuntu",
        "user",
    }

    def _rewrite(m: re.Match[str]) -> str:
        user = (
            m.group("user_name")
            if (m.group("user_name") in known_users or is_jinja2_expression(m.group("user_name")))
            else "ano-user"
        )
        return m.group("before") + user

    user_regexes = [
        r"(?P<before>[c-z]:\\users\\)(?P<user_name>(\w|{{\s*.*?\s*}}){,255})",
        r"(?P<before>/(home|Users)/)(?P<user_name>([a-z0-9_-]|{{\s*.*?\s*}}){,255})",
    ]
    for regex in user_regexes:
        block = re.sub(regex, _rewrite, block, flags=flags)
    return block


def hide_secrets(block: str, value_template: Template) -> str:
    root_node = parse_raw_block(block)

    output = ""
    for node in flatten(root_node):
        if node.type is NodeType.secret:
            if node.previous.type is node.holder:  # type: ignore[comparison-overlap]
                # Already quoted
                quote = ""
            else:
                quote = "" if node.holder and node.holder.text else '"'
            if not node.secret_value_of:
                # Should never happen
                continue
            output += quote + anonymize_field("", node.secret_value_of.text, value_template) + quote
        else:
            output += node.text
    return output


def anonymize_text_block(block: str, value_template: Optional[Template] = None) -> str:
    if not value_template:
        value_template = Template("{{ $variable_name }}")

    block = hide_comments(block)
    block = hide_secrets(block, value_template)
    block = hide_emails(block)
    block = hide_ip_addresses(block)
    block = hide_us_ssn(block)
    block = hide_mac_addresses(block)
    block = hide_us_phone_numbers(block)
    block = hide_credit_cards(block)
    block = hide_user_name(block)

    return block



#!/usr/bin/env python3
"""Jinja2 related function(s)."""
import re


def str_jinja2_variable_name(name: str) -> str:
    """Sanitize a string to make it suitable to become a Jinja2 variable."""
    name = name.replace("-", "_")
    name = re.sub(r"[^a-z_\d]", "", name, flags=re.IGNORECASE)
    return name.lower().lstrip("_")



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=invalid-name
"""Identify and merge the multilines nodes."""
from .node import Node, NodeType


class Line:
    """A wrapper class for a list of nodes that compose a line."""

    def __init__(self, space_indent_length: int, nodes: list[Node]) -> None:
        self.space_indent_length = space_indent_length
        self.content = ""
        for i in nodes:
            self.content += i.text
        self.nodes: list[Node] = nodes

    def get_last_node(self) -> Node:
        """Return the last node of the line."""
        return self.nodes[-1]


def is_beginning_of_multiline_block(node: Node) -> bool:
    if node.text not in ["|", ">"]:
        return False
    previous_nodes = list(reversed(list(node.get_previous_nodes(limit=3))))
    next_nodes = list(node.get_next_nodes(limit=2))
    if [n.type for n in previous_nodes] != [NodeType.field, NodeType.separator, NodeType.space]:
        return False
    if previous_nodes[1].text != ":":
        return False
    if [n.type for n in next_nodes] != [NodeType.new_line, NodeType.space]:
        return False
    return True


def read_one_line(node: Node) -> Line:
    space_indent_length: int = 0
    nodes: list[Node] = []

    c = node

    while c and c.next and c.type == NodeType.space:
        space_indent_length += len(c.text)
        nodes.append(c)
        c = c.next
    while c:
        nodes.append(c)
        if c.type == NodeType.new_line:
            break
        c = c.next
    return Line(space_indent_length, nodes)


def get_lines(node: Node) -> list[Line]:
    c = node.next  # Slip the first \n
    while c and c.next and c.type is NodeType.space:
        c = c.next
    if not (c and c.next):
        return []
    c = c.next
    first_line: Line = read_one_line(c)
    lines: list[Line] = [first_line]
    c = first_line.get_last_node()
    while c and c.next:
        new_line = read_one_line(c.next)
        if new_line.space_indent_length < first_line.space_indent_length:
            break
        lines.append(new_line)
        c = new_line.get_last_node()
    return lines


def group_multi_lines(node: Node) -> None:
    """Iterator that go through each nodes and follow the original text order."""
    current = node
    while current:
        if is_beginning_of_multiline_block(current):
            lines = get_lines(current)
            if not lines:
                current = current.next
                continue
            length = sum(len(i.nodes) for i in lines)
            for idx in range(length):
                if length == idx + 1 and current.next and current.next.type is NodeType.new_line:
                    break
                current.merge_with_next()
            current.type = NodeType.field
            # The if statement that should always be True because
            # is_beginning_of_multiline_block() was used first,
            # This is used for mypy/pyright to avoid the following error:
            # "previous" is not a known member of "None"
            if current.previous and current.previous.previous:
                current.secret_value_of = current.previous.previous.previous
        current = current.next



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=R0801
from ipaddress import IPv4Address, IPv4Network, IPv6Address
from string import Template
from textwrap import dedent

from ansible_anonymizer.anonymizer import (
    anonymize,
    anonymize_field,
    anonymize_struct,
    anonymize_text_block,
    hide_comments,
    hide_credit_cards,
    hide_emails,
    hide_ip_addresses,
    hide_mac_addresses,
    hide_us_phone_numbers,
    hide_us_ssn,
    hide_user_name,
    redact_ip_address,
    redact_ipv4_address,
    redact_ipv6_address,
    unquote,
)


def test_redact_ipv4_address():
    assert redact_ipv4_address(IPv4Address("192.168.3.5")) in IPv4Network("192.168.3.0/24")
    assert redact_ipv4_address(IPv4Address("8.8.8.8")) == IPv4Address("8.8.8.8")
    assert redact_ipv4_address(IPv4Address("8.8.8.9")) in IPv4Network("8.8.8.0/24")


def test_redact_ipv6_address():
    assert redact_ipv6_address(IPv6Address("2001:4860:4860::8888")) == IPv6Address(
        "2001:4860:4860::8888"
    )
    assert IPv6Address(redact_ipv6_address(IPv6Address("2001:db8:3333:4444:5555:6666:7777:8888")))


def test_redact_ip_address():
    assert redact_ip_address("2001:4860:4860::8888") == "2001:4860:4860::8888"
    assert IPv4Address(redact_ip_address("8.8.8.9"))


def test_anonymize_struct_nested_struct():
    in_ = {
        "name": "Install nginx and nodejs 12",
        "apt": {"name": ["nginx", "nodejs"], "state": "latest"},
        "a_set": {1, 2, 3},
        "dict_wit_with_int_as_index": {1: "1", 2: "2", 3: "3"},
    }
    assert anonymize_struct(in_) == in_


def test_anonymize_struct_ip_addresses():
    in_ = {
        "name": "foo@montreal.ca",
        "a_module": {
            "ip": ["2001:460:48::888", "192.168.1.1"],
            "password": "@This-should-disapear!",
        },
    }
    changed = anonymize_struct(in_)
    assert "foo@montreal.ca" not in changed["name"]
    assert "2001:460:48::888" not in changed["a_module"]["ip"]
    assert changed["a_module"]["password"] == "{{ password }}"


def test_anonymize_multiple_passwords():
    in_ = {"password": ["first_password", "second_password"]}
    assert anonymize_struct(in_) == {"password": ["{{ password }}", "{{ password }}"]}


def test_anonymize_email():
    # Str
    in_ = "my-email-address@somewhe.re"
    changed = anonymize_struct(in_)
    assert in_ not in changed
    assert isinstance(changed, str)
    assert "@" in changed

    # List
    in_ = ["my-email-address@somewhe.re"]
    changed = anonymize_struct(in_)
    assert in_ != changed
    assert isinstance(changed[0], str)
    assert "@" in changed[0]


def test_anonymize_with_special_template():
    original = {"password": ["first_password", "second_password"], "block": 'password1: "bar"'}
    expected = {"password": ["ö", "ö"], "block": 'password1: "ö"'}
    value_template = Template("ö")
    value_template = Template("ö")
    assert anonymize_struct(original, value_template=value_template) == expected


def test_anonymize_deprecated_function():
    in_ = ["my-email-address@somewhe.re"]
    changed = anonymize(in_)
    assert in_ != changed


def test_anonymize_text_block_no_change():
    source = """
    ---
    - name: AWS Cloud Operations
      hosts: localhost
      tasks:
        - name: Create a virtual network named myvpc
          amazon.aws.ec2_vpc_net:
            aws_access_key: "{{ aws_access_key }}"
            aws_secret_key: "{{ aws_secret_key }}"

    ---
    - name: AWS Cloud Operations
      hosts: localhost
      vars:
        myvpc_region: "us-east1"
        myvpc_name: "myvpc"

      tasks:
        - name: Create a virtual network
          amazon.aws.ec2_vpc_net:
            name: "{{ myvpc_name }}"
            cidr_block: "{{ myvpc_cidr_block }}"

    ---
    - name: Add mysshkey to Linux servers
      ansible.posix.authorized_key:
        user: "{{ user }}"
        state: present
        key: "{{ lookup('file', '/mysshkey') }}"

    """
    assert anonymize_text_block(dedent(source)) == dedent(source)


def test_anonymize_text_block_secret_fields():
    source = """

        - name: some example
            a-broken-key:
                my-secret: a-secret
                @^my-secret: weird-artifact
                %@iÜ-secret: "again"
                quoted-secret: ' With {{ some_variable }} again    '
                private_key: ~/.ssh/id_rsa

    """
    expectation = """

        - name: some example
            a-broken-key:
                my-secret: "{{ my_secret }}"
                @^my-secret: "{{ my_secret }}"
                %@iÜ-secret: "{{ secret }}"
                quoted-secret: '{{ quoted_secret }}'
                private_key: "{{ private_key }}"

    """
    assert anonymize_text_block(source) == expectation


def test_anonymize_text_block_email_addresses():
    source = """

        - name: some example
            a-broken-key:
                emails: - fooo@bar.ca
                - pierre-loup@some.company
                - christina@world.corp
                - "christina@world.corp"
                - 'christina@world.corp'

    """
    expectation = """

        - name: some example
            a-broken-key:
                emails: - lucas14@example.com
                - elijah6@example.com
                - evelyn17@example.com
                - "evelyn17@example.com"
                - 'evelyn17@example.com'

    """
    assert anonymize_text_block(source) == expectation
    assert hide_emails(source) == expectation


def test_anonymize_text_block_ip_addresses():
    source = """

        - name: some example
            this-should-remain: 8.8.8.8
            a-broken-key:
                some-random-ips: - fda4:597b:21fc:d31f::
                - 23.233.103.236
                - 192.168.10.34
                - 192.168.10.34/32
                - fda4:597b:21fc:d31f::/128
    """
    expectation = """

        - name: some example
            this-should-remain: 8.8.8.8
            a-broken-key:
                some-random-ips: - fda4:17b:1fc:31f::
                - 23.233.104.40
                - 192.168.10.48
                - 192.168.10.48/32
                - fda4:17b:1fc:31f::/128
    """
    assert anonymize_text_block(source) == expectation
    assert hide_ip_addresses(source) == expectation


def test_anonymize_text_block_us_ssn():
    source = """

    - copy:
        content: |
          here some content with a ssn "078-05-1120"
          and this is pi: 3.1415926535897936

    """

    expectation = """

    - copy:
        content: |
          here some content with a ssn "{{ ssn }}"
          and this is pi: 3.1415926535897936

    """
    assert anonymize_text_block(source) == expectation
    assert hide_us_ssn(source) == expectation


def test_anonymize_text_block_macaddress():
    source = """

    - copy:
        content: |
          some mac addresses "a0:36:9f:0e:9d:78"
          or A0-36-9F-0E-9D-78
          or A036.9F0E.9D78
          and this is pi: 3.1415926535897936

    """

    expectation = """

    - copy:
        content: |
          some mac addresses "5b:e1:4a:b9:48:23"
          or f5-8b-e4-53-e2-cd
          or 39cf.2897.2601
          and this is pi: 3.1415926535897936

    """
    assert anonymize_text_block(source) == expectation
    assert hide_mac_addresses(source) == expectation


def test_anonymize_text_block_us_phone_numbers():
    source = """

    - copy:
        content: |
            (914) 499-1900
            "914-499-1900"
            "914 499-1900"
            9144991900
            19144991900
          and this is pi: 3.14159265358
          a french number: 06 10 00 10 23
          a cell number from belgium: 0479 20 07 77

    """

    expectation = """

    - copy:
        content: |
            (311) 555-2368
            "(311) 555-2368"
            "(311) 555-2368"
            (311) 555-2368
            (311) 555-2368
          and this is pi: 3.14159265358
          a french number: 06 10 00 10 23
          a cell number from belgium: 0479 20 07 77

    """
    assert anonymize_text_block(source) == expectation
    assert hide_us_phone_numbers(source) == expectation
    assert hide_us_phone_numbers("914-499-1900") == "(311) 555-2368"


def test_anonymize_text_block_credit_cards():
    source = """

    - copy:
        content: |
          a_quoted_cc_number("1234567812345670")
          1234567812345670
          (1234567812345670)
          "1234567812345670"
          "1234 5678 1234 5670"
          "1234-5678-1234-5670"
          a UUID that look like CC number: "34206f73-4e3a-1234-567812345670-b85a"
          and this is pi: 3.1415926535897936

    """

    expectation = """

    - copy:
        content: |
          a_quoted_cc_number("{{ credit_card_number }}")
          {{ credit_card_number }}
          ({{ credit_card_number }})
          "{{ credit_card_number }}"
          "{{ credit_card_number }}"
          "{{ credit_card_number }}"
          a UUID that look like CC number: "34206f73-4e3a-1234-567812345670-b85a"
          and this is pi: 3.1415926535897936

    """
    assert anonymize_text_block(source) == expectation
    assert hide_credit_cards(source) == expectation
    assert hide_credit_cards("1234 5678 1234 5670") == "{{ credit_card_number }}"


def test_anonymize_text_block_comments():
    source = """

    # That a task block
    - copy:  # A comment at the end of line
        content: "some value to #  keep"

    """

    expectation = """


    - copy:
        content: "some value to #  keep"

    """
    assert anonymize_text_block(source) == expectation
    assert hide_comments(dedent(source)) == dedent(expectation)


def test_anonymize_comment_with_quote():
    source = (
        "# NEWCOPY'- "
        "#name: using the zos_operator module, issue command PHASEIN  "
        "#zos_operator:    cmd: 'CEMT SET PROGRAM PHASEIN'"
    )
    assert hide_comments(source) == ""


def test_anonymize_text_block_user_name():
    source = """
    "documentUri": "file:///home/pierre-yves/git_repos/ansible-collections/tag_operations.yml"
    'documentUri': 'file:///Users/rbobbitt/work//full_playbook.yml',
    "dest": "/home/fedora/somewhere-else"
    "dest": "/home/ubuntu/somewhere-else"
    "dest": "c:\\Users\\Gilbert\\été \\directory"
    "unicode": c:\\Users\\Eloïse\\Œufs de pâques\\Fête
    some_field:
      - /home/marie-pier
      - /home/marie-pier"Not the login"
      - c:\\Users\\Bảo
      - c:\\Users\\Bảo"Not the login"

    """

    expectation = """
    "documentUri": "file:///home/ano-user/git_repos/ansible-collections/tag_operations.yml"
    'documentUri': 'file:///Users/ano-user/work//full_playbook.yml',
    "dest": "/home/fedora/somewhere-else"
    "dest": "/home/ubuntu/somewhere-else"
    "dest": "c:\\Users\\ano-user\\été \\directory"
    "unicode": c:\\Users\\ano-user\\Œufs de pâques\\Fête
    some_field:
      - /home/ano-user
      - /home/ano-user"Not the login"
      - c:\\Users\\ano-user
      - c:\\Users\\ano-user"Not the login"

    """
    assert anonymize_text_block(source) == expectation
    assert hide_user_name(dedent(source)) == dedent(expectation)


def test_anonymize_text_block_username_for_linux_path():
    assert (
        anonymize_text_block("path: /home/kaisersoze/.ssh/authorized_keys")
        == "path: /home/ano-user/.ssh/authorized_keys"
    )


def test_anonymize_text_block_username_for_windows_path():
    assert (
        anonymize_text_block("path: C:\\users\\kaisersoze\\test")
        == "path: C:\\users\\ano-user\\test"
    )


def test_anonymize_text_block_username_as_jinja_template_for_linux_path():
    assert (
        anonymize_text_block(
            "path: /home/{{ admin_username | default('azureuser') }}/.ssh/authorized_keys"
        )
        == "path: /home/{{ admin_username | default('azureuser') }}/.ssh/authorized_keys"
    )


def test_anonymize_text_block_username_as_jinja_template_for_windows_path():
    assert (
        anonymize_text_block("path: c:\\users\\{{ admin_username | default('azureuser') }}\\test")
        == "path: c:\\users\\{{ admin_username | default('azureuser') }}\\test"
    )


def test_anonymize_field():
    field = "my_field"
    value = "     a    "

    value_template = Template("{{ $variable_name }}")
    assert anonymize_field(value, field, value_template) == value


def test_unquote():
    assert unquote("'a'") == "a"
    assert unquote('"a"') == "a"
    assert unquote('"a\'') == '"a\''
    assert unquote("a") == "a"
    assert unquote("''") == ""
    assert unquote("'") == "'"


def test_anonymize_uuid_field():
    field = "uuid_field"
    value = "ce34efc1-f5e3-4b0f-bb2c-5272319589a7"
    value_template = Template("{{ $variable_name }}")
    assert anonymize_field(value, field, value_template) == value

    value = "CE34EFC1-F5E3-4B0F-BB2C-5272319589A7"
    value_template = Template("{{ $variable_name }}")
    assert anonymize_field(value, field, value_template) == value


def test_anonymize_special_template():
    field = "secret"
    value = "to_hide"

    value_template = Template("--${variable_name}--")
    assert anonymize_field(value, field, value_template) == "--secret--"


def test_hide_secrets_aws_profile():
    origin = """
    [my-secret-account]
    aws_access_key_id = BJIA5UUFYYOOKZQODC3F
    aws_secret_access_key = NeTL/2vPPnlnb/8RBtsw3EwnNjflDbgZiDmRskhb
    """

    expectation = """
    [my-secret-account]
    aws_access_key_id = "{{ aws_access_key_id }}"
    aws_secret_access_key = "{{ aws_secret_access_key }}"
    """
    assert anonymize_text_block(dedent(origin)) == dedent(expectation)


def test_hide_secrets_multi_secrets():
    origin = """
    '(?i)password1:': "{{ _iosxr_password }}"
    "this is somethingpass: password2: else my_password3: 'password4: _Agaim': barfoo"
    password5: maxplus
    password6: "maxplus"
    password7: "my_password8: maxplus"
    "password9": "my_password10: maxplus"
    password11: "my_password12:
              maxplus"

    passwd: $6$j212wezy$7H/1LT4f9/N3wpgNunhsIqtMj62OKiS3nyNwuizouQc3u7MbYCarYeAHWYPYb2FT.lbioDm2RrkJPb9BZMN1O/
    """  # noqa: E501
    expectation = """
    '(?i)password1:': "{{ _iosxr_password }}"
    "this is somethingpass: {{ somethingpass }}: else my_password3: '{{ my_password3 }}': barfoo"
    password5: "{{ password5 }}"
    password6: "{{ password6 }}"
    password7: "{{ password7 }}"
    "password9": "{{ password9 }}"
    password11: "{{ password11 }}"

    passwd: "{{ passwd }}"
    """
    assert anonymize_text_block(origin) == expectation


def test_hide_secret_sudo_line():
    source = 'line="%wheel\tALL=(ALL)\tNOPASSWD: ALL"'
    assert anonymize_text_block(source) == source


def test_anonymize_text_block_quoted():
    assert (
        anonymize_text_block("ansible: 'ALL=(ALL) PASSWD: \\\"{{NOPASSWD'")
        == "ansible: 'ALL=(ALL) PASSWD: {{ passwd }}'"
    )
    assert (
        anonymize_text_block("ansible: 'ALL=(ALL) PASSWD: \\\"{{NOPASSWD'")
        == "ansible: 'ALL=(ALL) PASSWD: {{ passwd }}'"
    )
    assert (
        anonymize_text_block("password1: 'foobar'\npassword: 'barfoo'")
        == "password1: '{{ password1 }}'\npassword: '{{ password }}'"
    )
    assert (
        anonymize_text_block('%wheel	ALL=(ALL)	PASSWD: "ALL"')
        == '%wheel	ALL=(ALL)	PASSWD: "{{ passwd }}"'  # noqa: E501
    )


def test_anonymize_text_block_special_template():
    value_template = Template("--")
    assert anonymize_text_block('my_secret: "ÓÐG™ÉÓÖ"', value_template) == 'my_secret: "--"'


def test_anonymize_text_block_preserve_protected_quotes():
    assert (
        anonymize_text_block('line: "%ansible password=\'foobar\'"')
        == 'line: "%ansible password=\'{{ password }}\'"'
    )


def test_anonymize_multi_lines():
    origin = """
    foo: |
      line1
      line2
    my_secret: |
      line3
      line4
    """
    expectation = """
    foo: |
      line1
      line2
    my_secret: "{{ my_secret }}"
    """

    assert anonymize_text_block(origin) == expectation



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

from textwrap import dedent

from ansible_anonymizer.parser import (
    NodeType,
    breakup_elements,
    combinate_value_fields,
    flatten,
    parse_raw_block,
)


def test_hide_secrets_trailing_secret():
    sample = "password1: foobar\npassword: barfoo"
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {"password1": "foobar", "password": "barfoo"}


def test_hide_secrets_quoted_field():
    sample = """
    "password9": "my_password10: maxplus"
    """
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {"password9": "my_password10: maxplus"}


def test_parse_secrets_unquoted_field():
    sample = """
    password9: "my_password10: maxplus"
    """
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {"password9": "my_password10: maxplus"}


def test_parse_secrets_pattern_within_quoted_string():
    sample = """
    'password9: "my_password10: maxplus"'
    """
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {"password9": "my_password10: maxplus"}


def test_parse_secrets_long_string():
    sample = """
    passwd: $6$j212wezy$7H/1LT4f9/N3wpgNunhsIqtMj62OKiS3nyNwuizouQc3u7MbYCarYeAHWYPYb2FT.lbioDm2RrkJPb9BZMN1O/
    """  # noqa: E501
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {
        "passwd": (
            "$6$j212wezy$7H/1LT4f9/N3wpgNunhsIqtMj62OKiS3nyNwuizo"
            "uQc3u7MbYCarYeAHWYPYb2FT.lbioDm2RrkJPb9BZMN1O/"
        )
    }


def test_parse_secrets_2_level_of_quotes():
    sample = """
    "aaa'
       passwd: bob'"
    """
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {"passwd": "bob"}


def test_parse_secrets_protected_double_quotes():
    sample = """
    "aaa
       \\"passwd: bob\\""
    """
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {"passwd": "bob"}


def test_parse_secrets_empty():
    sample = ""
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert len(nodes) == 1


def test_parse_secrets_field_only():
    sample = "passwd"
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert len(nodes) == 2
    assert nodes[-1].text == sample


def test_parse_secrets_vars_file():
    sample = """
    ansible_user: root
    ansible_host: esxi1-gw.ws.testing.ansible.com
    ansible_password: '!234AaAa56'
    """
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {"ansible_password": "!234AaAa56"}


def test_parse_secrets_unquoted_string():
    sample = """
    ansible_password: an unquoted string
    """
    root_node = parse_raw_block(dedent(sample))
    nodes = list(flatten(root_node))
    assert nodes[-2].text == "an unquoted string"
    assert nodes[-2].secret_value_of.text == "ansible_password"


def test_parse_secrets_multi_secrets():
    sample = """
    '(?i)password1:': "{{ _iosxr_password }}"
    "this is somethingpass: password2: else my_password3: 'password4: _Agaim': barfoo"
    password5: maxplus
    password6: "maxplus"
    password7: "my_password8: maxplus"
    "password9": "my_password10: maxplus"
    password11: "my_password12:
              maxplus"

    passwd: $6$j212wezy$7H/1LT4f9/N3wpgNunhsIqtMj62OKiS3nyNwuizouQc3u7MbYCarYeAHWYPYb2FT.lbioDm2RrkJPb9BZMN1O/
    """  # noqa: E501
    root_node = parse_raw_block(sample)
    passwords = {
        t.secret_value_of.text: t.text for t in flatten(root_node) if t.type is NodeType.secret
    }
    assert passwords == {
        "somethingpass": "password2",
        "my_password3": "password4: _Agaim",
        "password5": "maxplus",
        "password6": "maxplus",
        "password7": "my_password8: maxplus",
        "password9": "my_password10: maxplus",
        "password11": "my_password12:\n              maxplus",
        "passwd": (
            "$6$j212wezy$7H/1LT4f9/N3wpgNunhsIqtMj62OKiS3n"
            "yNwuizouQc3u7MbYCarYeAHWYPYb2FT.lbioDm2RrkJPb9"
            "BZMN1O/"
        ),
    }


def test_parser_simple_key_value_string():
    sample = 'config_reverseproxy_oauth_password: "passw0rd"'
    expectation = [
        ("", NodeType.quoted_string_holder),
        ("config_reverseproxy_oauth_password", NodeType.field),
        (":", NodeType.separator),
        (" ", NodeType.space),
        ('"', NodeType.quoted_string_holder),
        ("passw0rd", NodeType.field),
        ('"', NodeType.quoted_string_closing),
    ]
    root_node = breakup_elements(sample)
    expanded = [(c.text, c.type) for c in flatten(root_node)]
    assert expanded == expectation


def test_parser_multi_spaces_before_simple_key_value_string():
    sample = 'config_reverseproxy_oauth_password:      "passw0rd"'
    expectation = [
        ("", NodeType.quoted_string_holder),
        ("config_reverseproxy_oauth_password", NodeType.field),
        (":", NodeType.separator),
        (" ", NodeType.space),
        (" ", NodeType.space),
        (" ", NodeType.space),
        (" ", NodeType.space),
        (" ", NodeType.space),
        (" ", NodeType.space),
        ('"', NodeType.quoted_string_holder),
        ("passw0rd", NodeType.field),
        ('"', NodeType.quoted_string_closing),
    ]
    root_node = breakup_elements(sample)
    expanded = [(c.text, c.type) for c in flatten(root_node)]
    assert expanded == expectation


def test_parser_get_secret():
    sample = "config_reverseproxy_oauth_password: my_secret"
    root_node = breakup_elements(sample)
    list_of_nodes = list(flatten(root_node))
    field_name_node = list_of_nodes[1]
    assert field_name_node.text == "config_reverseproxy_oauth_password"
    assert field_name_node.get_secret().text == "my_secret"


def test_parser_get_secret_with_unquoted_special_chars():
    sample = "config_reverseproxy_oauth_password: %$#my_secret&"
    root_node = breakup_elements(sample)
    field_name_node = root_node.next
    assert field_name_node.text == "config_reverseproxy_oauth_password"
    # Without combinate_value_fields we only get the first node of the secret
    assert field_name_node.get_secret().text == "%"
    combinate_value_fields(root_node)
    assert field_name_node.get_secret().text == "%$#my_secret&"


def test_parser_get_secret_with_ini_file():
    sample = """
    [default]
    foo = bar
    key=value
    turbo_secret=@#%$%^&^^ 645

    [section.bar]
    George = # a comment
    """
    root_node = breakup_elements(dedent(sample))
    secret_node = [n for n in flatten(root_node) if n.text == "turbo_secret"][0]
    # Without combinate_value_fields we only get the first node of the secret
    combinate_value_fields(root_node)
    assert secret_node.get_secret().text == "@#%$%^&^^ 645"


def test_combinate_value_fields():
    sample = "config_reverseproxy_oauth_password: my!secret%$!"
    root_node = breakup_elements(sample)
    assert len(list(flatten(root_node))) == 10
    combinate_value_fields(root_node)
    assert len(list(flatten(root_node))) == 5


def test_hide_secrets_multi_spaces_before_simple_key_value_string():
    sample = 'config_reverseproxy_oauth_password:      "passw0rd"'
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert nodes[-2].text == "passw0rd"
    assert nodes[-2].type is NodeType.secret


def test_parser_unquoted_password_with_special_chars():
    sample = 'my_password=      !pass w0rd"'
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert [t.type for t in nodes] == [
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.separator,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.secret,
    ]
    assert nodes[-1].type is NodeType.secret
    assert nodes[-1].text == '!pass w0rd"'


def test_parser_two_passwords_on_same_line():
    sample = 'my_password:      !pass w0rd  another_password: hide_thís "'
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert [t.type for t in nodes] == [
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.separator,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.secret,
        NodeType.space,
        NodeType.space,
        NodeType.field,
        NodeType.separator,
        NodeType.space,
        NodeType.secret,
    ]
    assert nodes[9].type is NodeType.secret
    assert nodes[9].text == "!pass w0rd"

    assert nodes[-1].type is NodeType.secret
    assert nodes[-1].text == 'hide_thís "'


def test_parser_one_password_and_one_regular_kv_on_same_line():
    sample = "my_password:      !pass w0rd  some-key: show-this"
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert [t.type for t in nodes] == [
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.separator,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.space,
        NodeType.secret,
        NodeType.space,
        NodeType.space,
        NodeType.field,
        NodeType.separator,
        NodeType.space,
        NodeType.field,
    ]
    assert nodes[9].type is NodeType.secret
    assert nodes[9].text == "!pass w0rd"


def test_parser_yaml_unquoted_password_with_a_equal_sign():
    sample = 'my_password:     !pass=w0rd"'
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert nodes[-1].secret_value_of is nodes[1]
    assert nodes[-1].type is NodeType.secret
    assert nodes[-1].text == '!pass=w0rd"'


def test_parser_ini_unquoted_password_with_a_colon_sign():
    sample = 'my_password = !pass:w0rd"'
    root_node = parse_raw_block(sample)
    nodes = list(flatten(root_node))
    assert nodes[-1].secret_value_of is nodes[1]
    assert nodes[-1].type is NodeType.secret
    assert nodes[-1].text == '!pass:w0rd"'


def test_parser_empty_quoted_secret():
    sample = 'password=""'
    root_node = breakup_elements(sample)
    node_types_found = [n.type for n in flatten(root_node)]
    assert node_types_found == [
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.separator,
        NodeType.quoted_string_holder,
        NodeType.quoted_string_closing,
    ]
    nodes_found = list(flatten(root_node))
    assert nodes_found[-2].closed_by == nodes_found[-1]


def test_parser_one_character_quoted_secret():
    sample = 'password="a"'
    root_node = breakup_elements(sample)
    node_types_found = [n.type for n in flatten(root_node)]
    assert node_types_found == [
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.separator,
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.quoted_string_closing,
    ]
    nodes_found = list(flatten(root_node))
    assert nodes_found[-3].closed_by == nodes_found[-1]



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

from ansible_anonymizer.node import NodeType
from ansible_anonymizer.parser import (
    breakup_elements,
    flatten,
)
from ansible_anonymizer.parser_multi_lines import (
    group_multi_lines,
)


def test_parser_multilines():
    sample = "a: |\n  my\n  multi\n  line\n"
    root_node = breakup_elements(sample)
    nodes_found_before = list(flatten(root_node))
    assert [n.type for n in nodes_found_before] == [
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.separator,
        NodeType.space,
        NodeType.unknown,
        NodeType.new_line,
        NodeType.space,
        NodeType.space,
        NodeType.field,
        NodeType.new_line,
        NodeType.space,
        NodeType.space,
        NodeType.field,
        NodeType.new_line,
        NodeType.space,
        NodeType.space,
        NodeType.field,
        NodeType.new_line,
    ]
    group_multi_lines(root_node)
    nodes_found_after = list(flatten(root_node))

    assert [
        NodeType.quoted_string_holder,
        NodeType.field,
        NodeType.separator,
        NodeType.space,
        NodeType.field,
        NodeType.new_line,
    ] == [n.type for n in nodes_found_after]
    assert nodes_found_after[-2].text == "|\n  my\n  multi\n  line"
    assert nodes_found_after[-1].text == "\n"


def test_parser_empty_multilines():
    sample = "a: |\nb: zz"
    root_node = breakup_elements(sample)
    nodes_found_before = list(flatten(root_node))
    group_multi_lines(root_node)
    nodes_found_after = list(flatten(root_node))
    assert [n.type for n in nodes_found_before] == [n.type for n in nodes_found_after]



#!/usr/bin/env python3
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

from ansible_anonymizer.field_checks import (
    is_allowed_password_field,
    is_jinja2_expression,
    is_password_field_name,
    is_path,
    is_uuid_string,
)


def test_is_allowed_password_field():
    assert is_allowed_password_field("NOPASSWD") is True
    assert is_allowed_password_field("NOPASSWD2") is False


def test_is_jinja2_expression():
    assert is_jinja2_expression("{{ foo|defaul('b')  }}") is True
    assert is_jinja2_expression("my_passw'rd") is False


def test_is_password_field_name():
    assert is_password_field_name("login") is False
    assert is_password_field_name("password") is True
    assert is_password_field_name("passwd") is True
    assert is_password_field_name("db_passwd") is True
    assert is_password_field_name("key_data") is True
    assert is_password_field_name("key_name") is True
    assert is_password_field_name("host_config_key") is True
    assert is_password_field_name("quayPassword") is True
    assert is_password_field_name("NOPASSWD") is False
    assert is_password_field_name("nopasswd") is True


def test_is_path():
    assert is_path("/etc/fstab") is True
    assert is_path("./opt/fstab") is True
    assert is_path("~/.ssh/id_rsa.pub") is True
    assert is_path(".%/mypassword/f$b") is False
    assert is_path("certificates/CA.key") is True
    assert is_path("a_password") is False


def test_is_uuid_string():
    assert is_uuid_string("ce34efc1-f5e3-4b0f-bb2c-5272319589a7") is True
    assert is_uuid_string("CE34EFC1-F5E3-4B0F-BB2C-5272319589A7") is True
